import numpy as np
class MicrophoneArray:
    """
    A class that represents a microphone array. The array is defined by its number of microphones
    and their positions

    Attributes
    ----------
    nmics: int
        number of microphones in the array
    mic_positions: np.ndarray
        2D array having dimensions `[nmics, 3]`, containing the `[x,y,z]` cartesian coordinates
        of each microphone in the array
    signals: np.ndarray
        2D array containing the `nimcs` signals acquired by each microphone in the array
    """

    def __init__(
            self,
            mic_positions,
        ) -> None:
        """
        Create microphone array object by defining the positions of the microphones contained the array

        Parameters
        ----------
        mic_positions: np.ndarray
            2D array having dimensions `[nmics, 3]`, containing the `[x,y,z]` cartesian coordinates
            of each microphone in the array

        Raises
        ------
        ValueError:
            If mic_positions is empty or contains less than 3 coordinates per microphone

        Modifies
        --------
        nmics
            The attribute is set equal to the first dimension of mic_positions (expressing the number of
            microhones in the array)
        """    
        if mic_positions.any() == None or np.shape(mic_positions)[0] == 0 or True in [len(i)!= 3 for i in mic_positions]:
            raise ValueError("mic_positions must contain at least one microphone. "
            "mic_positions must have shape [n, 3].")
        self.mic_positions = np.array(mic_positions)
        self.nmics = np.shape(mic_positions)[0]
        self.signals = None



import numpy as np

class SoundSource:
    """
    A class that defines a sound source. The sound source is assumed to be point-like and 
    omnidirectional, and emits an arbitrary sound signal. It is initially located in a 
    defined position, and moves along an arbitrary trajectory defined as a series of segments
    connecting a set of points, given as input.

    Attributes
    ----------
    position: np.ndarray
        1D array that contains initial position of sound source as a set of 3 cartesian coordinates `[x,y,z]`
    signal: np.ndarray
        1D array that contains samples of signal emitted by the sound source
    fs: int
        Sampling frequency of the emitted signal
    is_static: bool
        True if the source is static during the simulation
    static_simduration: int
        If the source is static, defines the duration of the simulations in seconds. If the source moves,
        the simulation duration is defined by the time it takes to travel the whole trajectory

    Methods
    -------
    set_trajectory(positions, speed):
        Defines a trajectory from a given set of N positions (`positions`) and N-1 velocities (`speed`). The speed
        is assumed to be constant between each couple of positions.
    set_signal(signal):
        Setter for signal attribute: assigns given signal to the sound source
    """

    def __init__(
            self,
            position = np.array([0,0,1]),
            fs = 8000,
            is_static = True,
            static_simduration = 5
        ) -> None:
        """
        Create SoundSource object by defining its initial position, trajectory, 
        emitted signal and sampling frequency.

        Parameters
        ----------
        position : ndarray
            1D Array containing 3 cartesian coordinates `[x,y,z]` that define initial source position, 
            by default [0,0,1]
        fs : int, optional
            Sampling frequency of the emitted signal, by default 8000
        is_static: Bool, optional
            True if the source is static
        static_simduration: float, optional
            If the source is static, defines the duration of the simulations in seconds. If the source moves,
            the simulation duration is defined by the time it takes to travel the whole trajectory
        """

        self.position = position
        self.signal = None
        self.fs = fs
        self.is_static = is_static
        self.static_simduration = static_simduration
        if self.is_static:
            # If duration is zero, consider single sample
            if self.static_simduration == 0:
                self.static_simduration = 1 / self.fs
            self.trajectory = np.tile(self.position, (round(self.fs * self.static_simduration), 1))
        
    def set_signal(self, signal: np.ndarray) -> None:
        """
        Setter for signal attribute: assigns given signal to the sound source

        Parameters
        ----------
        signal : np.ndarray
            1D Array containing samples of the source signal
        """
        self.signal = signal

    def set_trajectory(self, positions: np.ndarray, speed: np.ndarray) -> np.ndarray:
        """
        Defines a trajectory for the sound source from a set of N positions (given as triplets of Cartesian
        coordinates) and the values of the modulus of the source velocity between each subsequent couple of positions.
        The trajectory is defined as the positions assumed by the sound source at each sample of the simulation.
        Therefore, given the N input points, the method first computes a set of N-1 segments connecting one point to 
        the following one. Then, each segment is sampled according to the speed of the source in that segment and
        the signal sampling frequency, to yield a series of points so that each signal sample is emitted by the
        source at a different position of the trajectory.

        Parameters
        ----------
        positions : np.ndarray
            2D Array containing N sets of 3 cartesian coordinates `[x,y,z]` defining the desired trajectory positions.
            Each couple of subsequent points defines a straight segment on the overall trajectory
        speed : np.ndarray
            * 2D Array containing N-1 floats defining the modulus of the velocity on each trajectory segment
            * 1D Array containing one float, defining the modulus of the velocity on the whole trajectory (i.e. constant speed)

        Returns
        -------
        trajectory: np.ndarray
            2D Array containing N sets of 3 cartesian coordinates `[x,y,z]` defining the full sampled trajectory

        Raises
        ------
        ValueError
            If `speed` is neither a `float` nor a `np.ndarray` so that `len(speed) != (np.shape(positions)[0] - 1)`
        ValueError
            If `speed` is 0 or speed array contains value 0

        Modifies
        --------
        trajectory
            Parameter is updated with the new computed trajectory

        """

        if self.is_static:
            raise RuntimeError("Cannot assign trajectory to static source")

        trajectory = np.empty((0,3), dtype = np.float64)
        if len(speed) != (np.shape(positions)[0] - 1):
            if(len(speed) != 1):
                raise ValueError('Speed must be an array with len(speed) = np.shape(positions)[0] - 1 or len(speed) == 1!')
            # else:
                # Tile the speed value to cover the whole set of segments
                # speed = speed * np.ones(np.shape(positions)[0], 1)
                # speed = np.tile(speed, np.shape(positions)[0] - 1)
        if 0 in speed:
            raise ValueError("Speed cannot be zero")
        if len(speed) == 1:
            cum_lengths = np.zeros(len(positions) + 1)

            len_traj = 0
            b_seg = positions[0]
            

            for i in range(0, len(positions)):
                e_seg = positions[i]
                len_traj += np.sqrt((e_seg[0] - b_seg[0])**2 + (e_seg[1] - b_seg[1])**2)
                cum_lengths[i] = len_traj
                b_seg = e_seg
            
            cum_lengths[-1] = 2 * len_traj
            sim_time = len_traj / speed[0]

            num_points = int(sim_time * self.fs - 1)
            # trajectory = np.zeros((num_points,3))

            curr = 0
            next = 1

            for i in range(0, num_points):
                z = len_traj * i / num_points

                while(z > cum_lengths[next]):
                    curr += 1
                    next += 1
                if np.allclose(positions[curr], positions[next]):
                    curr += 1
                    next += 1
                
                b_seg = positions[curr]
                e_seg = positions[next]

                t = (z - cum_lengths[curr]) / (cum_lengths[next] - cum_lengths[curr])

                trajectory = np.append(trajectory, np.array([[(b_seg[0] * (1 - t) + e_seg[0] * t), b_seg[1] * (1 - t) + e_seg[1] * t, 1]]), axis = 0)

            # res[-1] = (seg_line[-1])
        else:
            for i in range(1, np.shape(positions)[0]):
                # Extremes of the considered segment
                a = positions[i - 1]
                b = positions[i]

                # Direction defining segment passing for a and b
                direction = b - a
                direction = direction / np.linalg.norm(direction)

                len_segment = np.sqrt(np.sum((a-b)**2))         # Compute length of segment A,B
                t_segment = len_segment / speed[i-1]            # Time to go from A to B (seconds)
                samples_segment = round(t_segment * self.fs)    # Number of samples to go from A to B

                # Positions on segment at each sample
                segment_positions = len_segment / samples_segment * range(samples_segment)
                # segment_positions = np.append(segment_positions, 1).reshape(-1,1)
                segment_positions = segment_positions.reshape(-1,1)
            
                segment_positions = np.tile(a,(len(segment_positions), 1)) + segment_positions * direction
                trajectory = np.append(trajectory, segment_positions, axis = 0)
            
        trajectory = np.append(trajectory, np.reshape(positions[-1], (1,3)), axis = 0)
        self.trajectory = trajectory

        return trajectory


import numpy as np
import math

class DelayLine:
    """
    Class that defines a variable length Delay Line. A delay line applies a delay of M samples
    to the input signal that is being written on the delay line. When the value of M varies at each
    sampling instant, it generates a continuously varying delay line, i.e. a delay line where the
    distance between the read pointer and the write pointer is a function of time.

    In order to take into account a smooth change in the delay length M, the read operation must rely
    on an interpolation, whenever the position of the read pointer falls between two samples of the delay
    line (i.e. M is a non-integer number, and the delay is thus fractional). Different interpolation
    methods (linear, allpass, sinc) result in different levels of accuracy.

    The variable length Delay Line can be used to simulate sound propagation if the delay varies according
    to the physical laws governing sound propagation.

    The delay line is represented by a circular `ndarray` of N samples, with a single write pointer performing
    write operations (one write per sampling interval), and an arbitrary number of read pointers performing
    interpolated read operations (one operation per sampling interval for each pointer).

    Attributes
    ----------
    N: int
        Number of samples (i.e. entries) of the array defining the delay line
    delay_line: np.ndarray
        Circular array of N samples (i.e. entries) defining the delay line
    write_ptr: int
        Position of the write pointer that performs write operations on the delay line. Value must be in `[0, N-1]`
    read_ptr: float or np.ndarray
        Position of read pointers that perform read operations on the delay line. Each pointer value must be in
        `[0, N-1]`, and can take float values (interpolated read operations are performed)
    interpolation: str
        Interpolation method to be used for fractional delay reads. Can be 'Linear', 'Allpass' or 'Sinc'
    

    Methods
    -------
    set_delays(delay):
        Set initial read pointers position, as specified by the `delay` parameter (delay in number of samples)
    update_delay_line(x, delay):
        Writes signal value x on the delay line at the write_ptr position and increments write_ptr by 1.
        Performs interpolated read operation for each of the write pointers, and updates read pointers values
        based on the delay parameter (delay in number of samples)

    Notes
    -----
    A detailed description of the working principle of variable length delay lines is given in 
    
    `https://ccrma.stanford.edu/~jos/pasp/Variable_Delay_Lines.html`
    
    """

    def __init__(
            self, 
            N = 48000,
            num_read_ptrs = 1,
            interpolation = 'Allpass',
        ):
        """
        Creates a Delay Line object as a `ndarray` of N samples, with one write pointer and M read pointers.

        Parameters
        ----------
        N : int, optional
            Number of samples (i.e. elements) of array defining the delay line, by default 48000
        num_read_ptrs : int, optional
            Number of pointers that will read values from the delay line, by default 1. The `write_ptr` attribute
            will be instantiated as an 1D `ndarray` having num_read_pointers float entries.
        interpolation : str, optional
            String describing the type of interpolator to be used for fractional delay read operations. Can be:
            * Linear: linear interpolation
            * Sinc: sinc interpolation with filter length of 11 taps
            * Allpass: first order all pass interpolator
            
            By default: 'Allpass'
        
        Raises
        ------
        ValueError:
            If `N` is negative or equal to zero
        ValueError:
            If `num_read_ptrs` <= 0
        ValueError:
            If `interpolation` is neither Linear, nor Sinc, nor Allpass
        """

        if N <= 0:
            raise ValueError("Delay Line size must be greater than zero.")
        if num_read_ptrs <= 0:
            raise ValueError("Number of read pointers must be greater than zero.")
        if (interpolation != 'Linear' and interpolation != "Lagrange" and
            interpolation != 'Allpass' and interpolation != "Sinc"):
            raise ValueError("Interpolation parameter can be: `Linear`, `Allpass` or `Sinc`")

        self.N = N
        self.write_ptr = 0
        self.read_ptr = np.zeros(num_read_ptrs, dtype=float)
        self.delay_line = np.zeros(N)
        self.interpolation = interpolation

        self._SINC_SMP = 11
        self._WINDOW = np.hanning(self._SINC_SMP)
        _table_size = 51
        self._delta = 1 / 50
        self._table_delays = np.arange(0,1+self._delta,self._delta)
        
        n = np.arange(self._SINC_SMP)
        self._sinc_table = np.zeros((_table_size, self._SINC_SMP))

        for idx, val in enumerate(self._table_delays):
            self._sinc_table[idx] = np.sinc(n - (self._SINC_SMP - 1) / 2 - val)
            self._sinc_table[idx] *= self._WINDOW
            self._sinc_table[idx] = self._sinc_table[idx] / np.sum(self._sinc_table[idx])

        self._ya_alt = np.zeros(len(self.read_ptr))    # Initial buffer parameter for all-pass interpolation

    def set_delays(self, delay: np.ndarray) -> None:
        """
        Sets initial delays between the write pointer and each of the read pointers.

        Parameters
        ----------
        delay : np.ndarray
            1D Array with number of elements equal to number of read pointers of the delay line. Each entry 
            is the delay in number of samples between the write pointer and each of read pointers, and can
            take non-integer positive values.

        Raises
        ------
        ValueError:
            If one or more elements of the delay array is zero or negative
        ValueError:
            If len(delay) is different from number of read pointers
        RuntimeError:
            If one or more elements of the delay array are greater than the length of the delay line
        """

        if np.any(delay <= 0):
            raise ValueError('Delays must be non-negative numbers')
        if len(delay) != len(self.read_ptr):
            raise ValueError('Length of delay array should match number of read pointers')
        if np.any(delay >= self.N):
            raise RuntimeError('Delay greater than delay line length has been encountered. Consider'
            'using a longer delay line')
        
        for i in range(len(self.read_ptr)):
            self.read_ptr[i] = self.write_ptr - delay[i]
            # Check that read pointer value is in [0, N-1]
            if (self.read_ptr[i] < 0):
                self.read_ptr[i] += self.N

    def update_delay_line(self, x: float, delay: np.ndarray) -> np.ndarray:
        """
        Writes signal value x on the delay line at the write_ptr position and increments write_ptr by 1.
        Performs an interpolated read operation for each of the read pointers, and updates read pointer values
        based on the corresponding delay parameter (delay in number of samples).

        Parameters
        ----------
        x : float
            Input signal sample to be written on delay line at write pointer position
        delay : np.ndarray
            1D Array with number of elements equal to number of read pointers of the delay line. Each entry 
            is the delay in number of samples between the write pointer and each of read pointers, and can
            take non-integer positive values.

        Returns
        -------
        np.ndarray
            1D Array containing 1 interpolated sample per each of the read pointers of the delay line. These 
            values represent the output of the delay line (i.e. a delayed version of the input signal, interpolated
            to take into account fractional values of the delay)
        """
        
        # Create array to store output values
        y = np.zeros(len(self.read_ptr))
        

        # Append input sample at the write pointer position and increment write pointer
        self.delay_line[self.write_ptr] = x
        self.write_ptr += 1

        for i in range(len(self.read_ptr)):
            # Compute interpolated read position (fractional delay)
            rpi = int(self.read_ptr[i])
            frac_del = self.read_ptr[i] - rpi
        
            # Produce output with interpolated read
            y[i] = self._interpolated_read(rpi, frac_del, self.interpolation, i)

            # Update read pointer position
            self.read_ptr[i] = self.write_ptr - delay[i]
            # Check that read and write pointers are within delay line length
            while (self.read_ptr[i] < 0):
                self.read_ptr[i] += self.N
            while (self.write_ptr >= self.N):
                self.write_ptr -= self.N
            while (self.read_ptr[i] >= self.N):
                self.read_ptr[i] -= self.N
        return y

    def _interpolated_read(self, read_ptr_integer: int, d: float, method: str = 'Allpass', rptr_idx: int = 0) -> float:
        """
        Performs interpolated reads from delay line. Read pointer position is given to this function splitted into an
        integer part (`read_ptr_integer`) and a fractional part (`d`) in [0,1]. The interpolation is 
        performed considering neighbouring samples of the signal on the delay line and using the method
        specified in the `method` parameter.
        
        Parameters
        ----------
        read_ptr_integer : int
            Integer part of the delay
        d : float
            Fractional part of the delay, in the interval [0,1]
        method : str, optional
            Interpolation method used, by default 'Allpass'. Can be:
            * Linear: linear interpolation, uses 2 neighbouring samples
            * Lagrange: Lagrange polynomial interpolation, order 5 can be changed in this function. Implemented
            using an FIR filter
            * Sinc: sinc interpolation, implemented using windowed truncated sinc FIR filter with 11 taps. Window
            and number of taps can be modified in this function
            * Allpass: first order all pass interpolator, uses 2 neighbouring samples
        rptr_idx: int
            Index of the read pointer that is currently being used to perform the read. Needed to store the value of
            the previous sample for each pointer, in case allpass filter is used

        Returns
        -------
        float
            Sample computed after interpolated read, as a single float value

        Raises
        ------
        ValueError:
            If `method` is neither `Linear`, nor `Lagrange`, nor `Sinc`, nor `Allpass`
        """
        
        # TODO Vectorization to speed up convolution

        if method == 'Lagrange':
            # Compute interpolation filter
            order = 5
            h_lagrange = self._frac_delay_lagrange(order, d)
            
            # Convolve signal with filter
            out = 0
            for i in range(0, len(h_lagrange)):
                out = out + h_lagrange[i] * self.delay_line[(read_ptr_integer + i - math.floor(order/2)) % self.N]
            return out

        elif method == 'Linear':
            # Linear Interpolation formula
            return d * self.delay_line[(read_ptr_integer + 1) % self.N] + (1 - d) * self.delay_line[read_ptr_integer]
        
        elif method == 'Allpass':
            y = self.delay_line[read_ptr_integer] + d * (self.delay_line[(read_ptr_integer + 1) % self.N] 
                - self._ya_alt[rptr_idx])
            self._ya_alt[rptr_idx] = y
            return y

        elif method == 'Sinc':
            # Define windowed sinc filter paramters and compute coefficients
            h_sinc = self._frac_delay_interpolated_sinc(d)
            
            # Convolve signal with filter
            out = 0
            for i in range(0, self._SINC_SMP):
                out = out + h_sinc[i]*self.delay_line[(read_ptr_integer + i - math.floor(self._SINC_SMP/2)) % self.N]
            return out

        else:
            raise ValueError("Interpolation parameter can be: `Linear`, `Lagrange`, `Allpass` or `Sinc`")
    
    def _frac_delay_lagrange(self, order: int, delay: float) -> np.ndarray:
        """
        Computes Lagrange fractional delay FIR filter coefficients.

        Parameters
        ----------
        order : int
            Filter order. Must be an odd number, number of taps will be order + 1
        delay : float
            Fractional part of the delay

        Returns
        -------
        np.ndarray
            1D Array containing (order + 1) FIR filter coefficients
        """

        n = np.arange(0, order + 1)
        h = np.ones(order + 1)

        for k in range(0, order + 1):
            # Find index n != k
            index = []
            for j in range(0, order + 1):
                if j != k:
                    index.append(j)
            
            h[index] = h[index] * (delay - k) / (n[index] - k)
        return h

    def _frac_delay_interpolated_sinc(self, delay: float) -> np.ndarray:
        """
        Computes windowed sinc FIR filter coefficients for sinc interpolation, using a table lookup and linear
        interpolation. In the initialization of the `DelayLine` component, a sinc table is computed, containing
        the samples of windowed sinc filters corresponding to a predefined set of fractional delays between 0 and 1,
        with a step size equal to `self._delta`. 

        Given the fractional part of the delay `delay`, this functions retrieves the two sinc arrays corresponding to 
        the nearest available delay points, and performs a linear interpolation to compute the windowed sinc samples
        corresponding to the delay `delay`.

        Parameters
        ----------
        delay : float
            Fractional part of the delay, in [0,1]

        Returns
        -------
        np.ndarray
            Windowed-sinc filter coefficients corresponding to the delay `delay`
        """
        alpha = delay % self._delta
        position = np.searchsorted(self._table_delays, delay - alpha)
        if alpha < 1e-9:
            return self._sinc_table[position]
        else:
            return alpha * (self._sinc_table[position] - self._sinc_table[position + 1]) + self._sinc_table[position + 1]

    def _frac_delay_sinc(self, sinc_samples: int, sinc_window: np.ndarray, delay: float) -> np.ndarray:
        """
        Computes windowed sinc FIR filter coefficients for sinc interpolation.

        Parameters
        ----------
        sinc_samples : int
            Number of taps of the filter, should be an odd number
        sinc_window : np.ndarray
            Window to be applied to sinc function. Length of the window must be equal to sinc_samples
        delay : float
            Fractional part of the delay

        Returns
        -------
        np.ndarray
            1D array containing `sinc_samples` FIR filter coefficients
        """
        sinc_filter = sinc_window * np.sinc(np.arange(0,sinc_samples) - (sinc_samples - 1) / 2 - delay)
        
        return sinc_filter / np.sum(sinc_filter)



import numpy as np
import math
import matplotlib.pyplot as plt

from .material import Material
from .soundSource import SoundSource
from .microphoneArray import MicrophoneArray
from .simulatorManager import SimulatorManager

class Environment:
    """
    This is the main class of the package and defines the Acoustic Scene that will be simulated. The goal of the
    package is the simulation of a moving sound source emitting sound above a flat surface. The sound propagation
    is simulated according to a model of the atmospheric sound propagation, and takes into account the sound
    that is being reflected by the road surface.

    The acoustic scene is composed of:
    * One sound source, emitting an arbitrary signal. It can be static or move along an arbitrary trajectory
    * One Microphone Array, containing a set of N omnidirectional microphones with an arbitrary disposition. The 
    array is static and is used to capture the sound produced by the moving source.
    * A set of environmental parameters describing the atmospheric conditions:
            * Atmospheric Pressure
            * Atmospheric Temperature
            * Relative Humidity
    * The material of the road surface
    * An optional diffuse background noise signal to be added to the simulation

    Alongside these parameters, simulation attributes are also defined:
    * The sampling frequency
    * The air absorption coefficients
    * The asphalt reflection coefficients

    From this class the simulation can be set up and run, via the instantiation of a SimulatorManager that
    contains the core functions of the simulator.

    Attributes
    ----------
    fs: int
        Sampling Frequency to be used for the simulations
    c: float
        Speed of sound in the air, computed from atmospheric parameters
    source: SoundSource
        Object containing parameters of the sound source
    background_noise: np.ndarray
        1D Array containing the samples of the background noise signal. The 
        background noise is assumed to be of diffuse type
    mic_array: MicrophoneArray
        Object containing the parameters of the microphone array (number and position of microphones)
    temperature: float
        Atmospheric temperature expressed in `Celsius` degrees
    pressure: float
        Atmospheric pressure expressed in `atm`
    rel_humidity: int
        Relative Humidity expressed as a percentage. Takes values between 0 and 100
    road_material: Material
        Object containing information about the absorption and reflection properties of the road surface
    air_absorption_coefficients: np.ndarray
        1D Array containing the air absorption coefficients computed at a set of 50 equispaced frequencies
        between 0 and `fs/2`
    Z0: float
        Value for air impedance computed at temperature `temperature`, pressure `pressure` and with 
        corresponding speed of sound `c`

    Methods
    -------
    set_road_material(absorption):
        Assign to the scene the material of the road surface, by selecting it from a database or creating a new
        material based on its acoustic absorption properties
    set_simulation_params(interp_method, include_reflection, include_air_absorption):
        Sets the parameters to be used in the simulation:
        * `interp_method`: interpolation method used to perform interpolated reads from delay line
        * `include_reflection`: if `False` only direct sound is simulated, if `True` also reflected path is included
        * `include_air_absorption`: if `True` effect of air absorption is included, if `False` it is neglected
    add_source(position, signal = None, trajectory = None):
        Creates a `SoundSource` by defining its initial position, its signal and (if it is a moving source) its
        trajectory, and inserts it into the acoustic scene
    set_background_noise(signal, SNR):
        Generates a background noise signal to be added to the signals generated at the microphone positions during
        the simulation. The noise level is defined by setting its SNR with respect to the source signal received by
        the microphone after the simulation is complete
    add_microphone_array(mic_locs):
        Creates a `MicrophoneArray` object by defining the positions of its microphones, and inserts it into the
        acoustic scene
    plot_environment():
        Plots the position of the microphones in the `MicrophoneArray` object and the trajectory of the `SoundSource`
    simulate():
        Runs the simulation and retrieves the signal received at each microphone in the `MicrophoneArray`

    """

    def __init__(
            self,
            fs: int = 8000,
            temperature: float = 20,
            pressure: float = 1,
            rel_humidity: int = 50,
            road_material: Material = Material('average_asphalt'),

        ) -> None:
        """
        Creates an Environment object by setting the simulation scene parameters. The Environment is defined by
        * a `SoundSource` object, representing a sound source moving along a trajectory
        * a `MicrophoneArray` object, representing a set of N microphones in fixed positions
        * a set of parameters describing the atmospheric conditions:
                * Temperature
                * Pressure
                * Relative Humidity
        * a `Material` object describing the absorption and reflection properties of the road surface
        * a sampling frequency `fs` used in the simulations

        Parameters
        ----------
        fs : int, optional
            Sampling frequency used in the simulations, by default 8000
        temperature : float, optional
            Atmospheric temperature in `Celsius` degrees, by default 20 degrees
        pressure : float, optional
            Atmospheric pressure in `atm`, by default 1 atm
        rel_humidity : int, optional
            Relative Humidity expressed as a percentage. Takes values in [0, 100], by default 50
        road_material : Material, optional
            Material object containing absorption and reflection properties of road surface, by default None. If 
            is None, an average asphalt material is chosen from Database.
        
        Raises
        ------
        ValueError:
            if rel_humidity is lower than zero or greater than 100
        """

        self.fs = fs
        self.source = None
        self.mic_array = None
        self.temperature = temperature
        self.pressure = pressure
        if rel_humidity < 0 or rel_humidity > 100:
            raise ValueError("Humidity must be greater than zero and lower than 100")
        self.rel_humidity = rel_humidity
        
        if road_material.absorption["center_freqs"][-1] != self.fs / 2:
            road_material.extrapolate_coeffs_to_spectrum(fs = self.fs)
        self.road_material = road_material

        simulation_params: dict = {
                "interp_method": "Sinc",
                "include_reflected_path": True,
                "include_air_absorption": True,
        }
        self.simulation_params = simulation_params
        
        self.c = self._compute_speed_sound(temperature)
        self.air_absorption_coefficients = self._compute_air_absorption_coefficients()
        self.Z0 = self._compute_air_impedance(self.temperature, self.pressure, self.c)

        # If true, background noise will be introduced
        self._background_noise_flag = False
        self._background_noise_SNR = 0
        self._background_noise = None
    
    def set_road_material(self, absorption: str or dict) -> None:
        """
        Set the road material by creating a new Material object from the absorption parameter. This parameter
        can either be a str containing a material name(the material will be chosen from the database based on 
        its name) or a dict containing keys `"coeffs"` and `"freqs"` to define the absorption properties of 
        an arbitrary material. See documentation of class `Material` for more information.

        Parameters
        ----------
        absorption : str or dict
            * If `str` is given: chooses `Material` from database
            * If `dict` is given: defines new `Material` having the `"coeffs"` and `"freqs"` specified in the
            dictionary.

        Modifies
        --------
        material
            The chosen or created material is assigned to the corresponding `Environment` attribute
        """

        material = Material(absorption)
        
        # If frequencies do not cover the whole spectrum, perform extrapolation
        if material.absorption["center_freqs"][-1] != self.fs / 2:
            material.extrapolate_coeffs_to_spectrum(fs = self.fs)
        
        self.road_material = material
    
    def set_simulation_params(self, interp_method: str = 'Allpass', include_reflection: bool = True, 
        include_air_absorption: bool = True) -> None:
        """
        Function used to define a set of parameters to be used in the simulation. These parameters are:
        * The interpolation method used for the interpolated reads from the delay lines. `Sinc` provides best
        accuracy but has highest computational load.
        * The possibility to include the reflected sound path in the simulation. If it is not included, the source is
        assumed to emit sound in the free-field and just the direct path is simulated: this reduces the computational
        load but also the physical accuracy
        * The possibility to include the air absorption in the simulation. The air absorption depends on distance and is
        implemented as a Time Varying FIR filter. The computation of its coefficients is performed at each simulation
        instant, and implies an increase of the computational complexity. This can be neglected, leading to a reduction in the
        complexity but also in the simulation accuracy.

        Parameters
        ----------
        interp_method : str
            Interpolation method used to perform interpolated reads from the delay lines. Can be:
            * `Linear`: linear interpolation
            * `Allpass` (default): first order allpass interpolation
            * `Sinc`: sinc interpolation using a windowed sinc filter with 11 taps. Highest accuracy and
            computational complexity 
        include_reflection : bool
            Bool, if `True` (default) the simulation includes the direct sound path and the road
            surface reflection. If `False` just the direct sound is simulated.
        include_air_absorption : bool
            Bool, if `True` (default), the air absorption is included in the simulation. The air absorption is modeled 
            using a FIR filter with 11 taps, whose coefficients depend on the distance between source and receiver 
            and thus are updated at each simulation instant. To reduce the computational load, this can be switched off 
            by setting this to `False`.
        """
        simulation_params = {
                "interp_method": interp_method,
                "include_reflected_path": include_reflection,
                "include_air_absorption": include_air_absorption,
            }
        self.simulation_params = simulation_params
    
    def add_source(self, position: np.ndarray, signal: np.ndarray = None, 
        trajectory_points: np.ndarray = None, source_velocity: np.ndarray or float = None) -> None:
        """
        Creates a sound source and adds it to the environment. To create the `SoundSource` object, 
        the specified parameters are required

        Parameters
        ----------
        position : np.ndarray
            Initial source position, specified as a 1D array containing three cartesian coordinates [x,y,z]
        signal : np.ndarray, optional
            Signal emitted by the sound source, by default None. If the signal is None, a default frequency 
            modulated sinusoidal signal is assigned to the source. The signal is looped to cover the whole 
            duration of the simulation
        trajectory_points : np.ndarray, optional
            2D Array containing N sets of 3 cartesian coordinates `[x,y,z]` defining the desired trajectory positions.
            Each couple of subsequent points define a straight segment on the overall trajectory. If value is None,
            the source is assumed to be static
        source_velocity : np.ndarray or float, optional
            * 2D Array containing N-1 floats defining the modulus of the velocity on each trajectory segment
            * float defining the modulus of the velocity on the whole trajectory (i.e. constant speed)

            If source is moving, it must be assigned. If source is static, it takes default value `None`

        Raises
        ------
        RuntimeError:
            If a source is already present in the acoustic scene (self.source != None)
        
        Modifies
        --------
        source
            The created `SoundSource` object is assigned to the corresponding attribute
        """
        #TODO: Normalize source signal
        if self.source is not None:
            raise RuntimeError("Cannot insert more than one sound source")
        is_static = False
        if trajectory_points is None:
            is_static = True
            source = SoundSource(position, self.fs, is_static)
        else:
            source = SoundSource(position = position, fs = self.fs, is_static = False)
            source.set_trajectory(trajectory_points, source_velocity)

        if signal is None:
            # Define a default signal --> sinusoidal siren
            simulation_duration = len(source.trajectory) / self.fs
            
            ## Define sinusoid with frequency modulation
            f = 800
            f_lfo = 0.3

            t = np.arange(0, simulation_duration, 1/self.fs)
            signal = np.zeros_like(t)
            for i in range(len(t)):
                signal[i] = 0.2 * np.sin(2 * np.pi * f * t[i] + 600 * np.sin(2 * np.pi * f_lfo * t[i]))

        # If trajectory is longer than signal, loop the signal
        while len(signal) < len(source.trajectory):
            signal = np.append(signal, signal)
        if len(signal) > len(source.trajectory):
            signal = signal[0:len(source.trajectory)]
        
        # Add signal to the sound source
        source.set_signal(signal)

        self.source = source
    
    def _add_noise_source(self, position: np.ndarray, signal: np.ndarray = None) -> None:
        """
        Creates a static noise signal and adds it to the acoustic scene in a specified position. Still
        to be implemented.

        Parameters
        ----------
        position : np.ndarray
            Noise source position, specified as a 1D array containing three cartesian coordinates [x,y,z]
        signal : [type], optional
            Signal emitted by the sound source, by default None. If the signal is None, a default white  
            noise signal is assigned to the source. The signal is looped to cover the whole  duration of 
            the simulation

        Raises
        ------
        NotImplementedError
            Function is still to be implemented
        """

        raise NotImplementedError('Noise Sources are not supported yet')
    
    def set_background_noise(self, signal: np.ndarray = None, SNR: float = 15) -> None:
        """
        Enables the use of a static background noise in the simulation. The noise is assumed to be of 
        diffuse type, and is defined by its `SNR`. This is computed as the SNR between the source signal
        received by the microphone after the simulation is complete, and the noise signal chosen by the 
        user. If no noise signal is chosen, an static white noise is used.

        Parameters
        ----------
        signal : np.ndarray, optional
            1D ndarray containing the samples of the background noise signal to be used during the simulation, 
            by default None
        SNR : float, optional
            Signal to Noise ratio in dB scale, by default 15

        Raises
        ------
        RuntimeError
            If the source or the mic_array has not been instantiated yet
        """

        if self.source == None or self.mic_array == None:
            raise RuntimeError("To add a background noise you need to first insert a sound source"
                "and a microphone array")

        self._background_noise_flag = True
        self._background_noise_SNR = SNR
        
        if signal is None:
            # Define default signal --> white noise
            simulation_duration = len(self.source.trajectory) / self.fs
            t = np.arange(0, simulation_duration, 1 / self.fs)
            signal = np.random.randn(len(t))
        
        else:
            while(len(signal) < len(self.source.trajectory)):
                signal = np.append(signal, signal)
            if len(signal) > len(self.source.trajectory):
                signal = signal[0:len(self.source.trajectory)]
        
        self._background_noise = signal

    def add_microphone_array(self, mic_locs: np.ndarray) -> None:
        """
        Creates a MicrophoneArray object and adds it to the Environment. The array is defined by the 
        position of its microphones.

        Parameters
        ----------
        mic_locs : np.ndarray
            2D Array containing N triplets of cartesian coordinates [x,y,z] defining the position of the 
            microphones of the array
        
        Raises
        ------
        RuntimeError:
            If a microphone array is already present in the acoustic scene when method is called

        Modifies
        --------
        mic_array
            The instantiated MicrophoneArray is assigned to the corresponding Environment attribute
        """
        if self.mic_array != None:
            raise RuntimeError("Cannot insert more than one microphone array")
        mic_array = MicrophoneArray(mic_locs)
        self.mic_array = mic_array

    def plot_environment(self):
        """
        Plots a 2D representation of the position of the microphones in `mic_array` and the trajectory of the 
        sound source during the simulation.
        """

        plt.figure()
        plt.plot(self.source.trajectory[:,0], self.source.trajectory[:,1])
        plt.scatter(self.mic_array.mic_positions[:,0], self.mic_array.mic_positions[:,1], color = 'green', marker='x')
        plt.legend(['Source Trajectory', 'Microphones'])
        plt.xlabel('[m]')
        plt.ylabel('[m]')
        plt.suptitle('Simulation Scenario')
        plt.title('Source and Microphones have height: z = %.2f' %self.mic_array.mic_positions[0,2])
        plt.show()

    def simulate(self) -> np.ndarray:
        """
        Runs the simulation and returns received microphone signals. To be called, the `Environment` must be fully
        specified, i.e. it must contain a `SoundSource` and a `MicrophoneArray`. The function relies on the
        `SimulatorManager`, a class that contains the functions to compute the new
        simulator output samples (i.e. the signals received at the microphones) and to update the acoustic scene
        at each simulation frame.

        Returns
        -------
        np.ndarray
            2D Array containing N samples of the signals received by each of the M microphones in the defined 
            microphone array.

        Raises
        ------
        RuntimeError
            If Source or MicrophoneArray have not been defined yet
        """
        
        if self.source == None or self.mic_array == None:
            raise RuntimeError('Before running simulation, you must define a sound source and a microphone array')
        
        # Duration of the simulation
        N = len(self.source.trajectory)
        M = self.mic_array.nmics

        signals = np.zeros((M,N))

        for m in range(M):
            # Select Active Microphone
            active_mic = self.mic_array.mic_positions[m]
        
            # Instantiate Simulator Manager
            manager = SimulatorManager(c = self.c, fs = self.fs, Z0 = self.Z0, road_material = self.road_material,
                airAbsorptionCoefficients = self.air_absorption_coefficients, simulation_params = self.simulation_params)

            # Define simulation loop
            manager.initialize(self.source.position, active_mic)

            # Compute output samples
            for n in range(N):
                signals[m,n] = manager.update(self.source.trajectory[n], active_mic, self.source.signal[n])
            

            # Add background noise
            if self._background_noise_flag == True:
                # Compute attenuation factor from SNR
                noise_attenuation = np.sqrt(np.sum(signals[m]** 2) / (10 ** (self._background_noise_SNR/10) 
                    * np.sum(self._background_noise ** 2)))
                
                # Compute background noise signal
                noise = self._background_noise * noise_attenuation
                
                # Compute signal
                signals[m] = signals[m] + noise
        return signals
    
    def _compute_air_absorption_coefficients(self, nbands: int = 20) -> np.ndarray:
        """
        Computes air absorption coefficients at a set of nbands equispaced frequencies in the range `[0, fs]`, based
        on the ISO 9613-1 standard. The coefficients depend on the atmsopsheric temperature, pressure and relative humidity.

        Parameters
        ----------
        nbands: int
            Number of frequency bands in which to compute air absorption coefficients, by default 20

        Returns
        -------
        np.ndarray
            1D Array containing nbands air absorption coefficients, computed in dB scale, at the defined 
            set of frequencies

        """

        T0 = 293.15     # Standard room temperature T = 20deg Celsius
        T01 = 273.16    # Triple point isotherm temperature, ISO 9613-1
        T = self.temperature + 273.15
        ps0 = 1         # Standard atmospheric pressure in atm

        f = np.linspace(0, self.fs / 2, num=nbands)     # Frequencies in which coeffs will be computed
        
        Csat = -6.8346 * math.pow(T01 / T, 1.261) + 4.6151
        rhosat = math.pow(10, Csat)
        H = rhosat * self.rel_humidity * ps0 / self.pressure

        frn = (self.pressure / ps0) * math.pow(T0 / T, 0.5) * (
                9 + 280 * H * math.exp(-4.17 * (math.pow(T0 / T, 1/3.) - 1)))

        fro = (self.pressure / ps0) * (24.0 + 4.04e4 * H * (0.02 + H) / (0.391 + H))

        alpha = f * f * (
            1.84e-11 / ( math.pow(T0 / T, 0.5) * self.pressure / ps0 )
            + math.pow(T / T0, -2.5)
            * (
                0.10680 * math.exp(-3352 / T) * frn / (f * f + frn * frn)
                + 0.01278 * math.exp(-2239.1 / T) * fro / (f * f + fro * fro)
                )
            )
        
        return alpha * 20 / np.log(10)
    
    def _compute_air_impedance(self, T: float, p: float = 1, c: float = None) -> float:
        """
        Compute Specific impedance of air, given temperature (in Celsius) and pressure (in atm). Speed of 
        sound can be given as a parameter; if None it is computed from temperature parameter.

        Parameters
        ----------
        T : float
            Atmospheric temperature in `Celsius` degrees
        p : float, optional
            Atmospheric pressure in `atm`, by default 1 atm
        c : float, optional
            Speed of sound in the air at T temperature, by default None. It can be given to speed up computation,
            otherwise it is computed from T inside the method

        Returns
        -------
        float
            Specific impedance Z0 of air

        """
        
        if c == None:
            c = self._compute_speed_sound(T)
        
        p = p * 101325          # Convert atm to Pascal
        T = T + 273.15          # Convert temperature to Kelvin
        R_spec = 287.058
        
        return p / (R_spec * T) * c

    def _compute_speed_sound(self, T: float) -> float:
        """
        Computes the speed of sound in the air at temperature T (expressed in Celsius degrees)

        Parameters
        ----------
        T : float
            Atmospheric Temperature expressed in Celsius degrees

        Returns
        -------
        float
            Value of speed of sound in air at temperature T

        """

        return 331.3 * np.sqrt(1 + T / 273.15)



import numpy as np
import matplotlib.pyplot as plt
import os, json, io


# Load materials data from json database and store in dict
_materials_database_fn = os.path.join(os.path.dirname(__file__), "materials.json")

with io.open(_materials_database_fn) as database_materials:
    materials_dict = json.load(database_materials)


class Material:
    """
    A class that describes the absorption properties of surfaces.

    Attributes
    ----------
    absorption: dict
        A dictionary containing keys ``description``, ``coeffs`` and ``center_freqs``
    reflection_coeffs: np.ndarray
        An array containing the reflection coefficients of the material in each frequency 
        band defined in the `absorption["center_freqs"]` attribute. Reflection coefficients
        are computed as:

        `refl_coeff[i] = np.sqrt(1 - abs_coeff[i])`
    
    Methods
    -------
    extrapolate_coeffs_to_spectrum(interp_degree = 2, fs = 8000, n_bands = 8):
        Extrapolate absorption coefficients to fill the frequency spectrum from 0 to fs/2
    plot_absorption_coeffs():
        Plot the absorption coefficients in the frequency bands considered in the analysis
    """
    
    def __init__(
            self,
            absorption,
        ) -> None:
        """
        Create the Material object by defining its absorption and reflection properties.

        Parameters
        ----------
        absorption: str or dict
        * str: the absorption coefficients will be obtained from the database
        * dict: a dictionary containing keys ``description``, ``coeffs`` and ``center_freqs``

        Raises
        ------
        KeyError
            If absorption is `dict` but does contain `coeffs` and `center_freqs` keys
        ValueError
            If number of elements in `coeffs` and `center_freqs` is different
        TypeError
            If absorption is neither `dict` nor `str`

        Modifies
        --------
        absorption
            Attribute gets a value depending on input parameter
        reflection_coeffs
            Compute reflection coefficients based on absorption coefficients value
        """

        if isinstance(absorption, str):
            absorption = dict(materials_dict[absorption])
        
        elif isinstance(absorption, dict):
            if "coeffs" not in absorption or "center_freqs" not in absorption:
                raise KeyError("'absorption' must be a dictionary with keys 'coeffs' and 'center_freqs'")

            if len(absorption["coeffs"]) != len(absorption["center_freqs"]):
                raise ValueError("Length of 'absorption['coeffs']' and "
                    "'absorption['center_freqs']' must match.")
        else:
            raise TypeError('Wrong Material Assignment')
        
        self.absorption = absorption
        self.reflection_coeffs = np.sqrt(1 - np.array(self.absorption["coeffs"]))
    
    def extrapolate_coeffs_to_spectrum(self, interp_degree: int = 2, fs: int = 8000, n_bands: int = 8) -> np.ndarray:
        """
        Extrapolate absorption coefficients to fill the frequency spectrum from 0 to fs/2.
        The extrapolation is operated via fitting a polynomial of degree interp_degree to
        the set of absorption coefficients obtained from the database or measured, usually 
        covering a subset of (low) frequencies of the full spectrum. 
        
        Returns a vector of n_bands absorption coefficients. The coefficients correspond to
        frequency bands, with center frequencies equispaced in the range [0, fs/2].

        Parameters
        ----------
        interp_degree : int, optional
            Degree of polynomial used for interpolation, by default 2
        fs : int, optional
            Sampling frequency - used to define frequency spectrum, by default 8000
        n_bands : int, optional
            Number of bands of the spectrum from 0 Hz to fs/2 Hz, by default 8. Each band
            is defined by its center frequency, and a single absorption coefficient (absorption
            properties are assumed to be constant within the same band).

        Returns
        -------
        estimated_abs_coeffs: np.ndarray
            Array with n_bands entries, containing absorption coefficients for frequency range [0, fs/2] Hz
        
        Modifies
        --------
        absorption
            Parameter gets new extrapolated Absorption coefficients 
        reflection_coeffs
            Parameter gets new reflection coefficients computed from extrapolated absorption coefficients
        """
        
        coeffs = self.absorption["coeffs"]
        freqs = self.absorption["center_freqs"]
        
        # Transform in log-domain to perform interpolation
        log_freqs = np.log(freqs)
        log_coeffs = np.log(coeffs)
        log_model = np.polyfit(log_freqs, log_coeffs, interp_degree)
        
        # Define new frequency spectrum
        full_spectrum = np.linspace(0, 1, n_bands) * fs / 2
        full_spectrum[0] = 0.01   # Introduce small variation on first coeffs to perform polyval
        
        # Compute interpolated coeffs
        log_estimated_abs_coeffs = np.polyval(log_model, np.log(full_spectrum))
        estimated_abs_coeffs = np.exp(log_estimated_abs_coeffs)
        full_spectrum[0] = 0   # Restore original full_spectrum

        # Clip between 0 and 1
        estimated_abs_coeffs[estimated_abs_coeffs < 0.0] = 0.0
        estimated_abs_coeffs[estimated_abs_coeffs > 1.0] = 1.0

        self.absorption = {
            "description":  self.absorption["description"],
            "coeffs": estimated_abs_coeffs,
            "center_freqs": full_spectrum
        }
        
        # Compute reflection coefficients from absorption
        self.reflection_coeffs = np.sqrt(1 - estimated_abs_coeffs)

        return estimated_abs_coeffs

    def plot_absorption_coeffs(self) -> None:
        '''
        Plot the frequency values of the absorption coefficients of the material.
        '''

        plt.figure()
        plt.plot(self.absorption["center_freqs"], self.absorption["coeffs"])
        plt.title('Absorption Coefficients Road Surface')
        plt.ylabel(r'$\alpha$')
        plt.xlabel('f [Hz]')
        plt.show()


import numpy as np
import scipy.signal
import math
from collections import deque

from .delayLine import DelayLine
from .material import Material

class SimulatorManager:
    """
    This class defines the simulation logic. It contains all the methods and components that are needed
    to run the acoustic simulation in the defined scenario, and compute the output signals. The
    acoustic scene defined in the `Environment` module contains a single sound source, an array of microphones, 
    and some additional parameters (e.g. a sampling frequency, the atmospheric conditions, the road surface properties,
    a background noise). The `SimulatorManager` simulates the propagation of sound from the moving source to a 
    single, static microphone, exploiting the scene parameters that it receives at the time of its initialization.

    The received signal will contain two different components: the direct sound, propagating directly from 
    the source (assumed to be omnidirectional and point-like) to the microphone (assumed
    to be omnidirectional and point-like too), and the sound reflected by the road surface. In order to simulate the
    reflection, the simulator exploits the Image-Source Method (ISM): a virtual source is generated by reflecting the
    position of the real sound source w.r.t the road surface, and the reflected sound is assumed to be emitted by
    this source as if it was emitting in the free-field. The sound produced the image source is then filtered by a 
    set of FIR filters that simulate the asphalt surface absorption and the attenuation due to the propagation of
    the sound through the air.
    
    The simulation is divided in two separate phases:
    * During the `initialization` phase, the `DelayLine` components used to simulate the sound propagation delay
    are instantiated, and the initial delays between the microphone, the sound source and the reflection point are
    computed.
    * During the `update` phase, first the signal emitted by the source is written on the two `DelayLine` instances 
    controlling the direct and reflected sound propagation delays. Then, the delays are updated according to the 
    new position of the moving source, together with the distances source -> receiver, image source -> road surface 
    and road surface -> receiver. The corresponding delayed signal samples are read from the `DelayLines` and filtered
    via a set of filters simulating the (angle-dependent) asphalt reflection and the air absorption. Finally, the
    signal received by the microphone is computed by summing the direct and the reflected components of the simulated
    sound field.

    Attributes
    -------
    c: float
        Speed of sound in the air
    fs: int
        Sampling frequency used in the simulations
    Z0: float
        Acoustic impedance of air
    road_material: Material
        Asphalt type of the road surface, defined by its absorption and reflection properties
    primaryDelayLine: DelayLine
        Primary `DelayLine` component, used to compute delays corresponding to propagation of sound between source
        and receiver on direct path, and between image source and road surface, on reflected path
    secondaryDelayLine: DelayLine
        Secondary `DelayLine` component, used to compute delay corresponding to the propagation of sound on the 
        reflected path from the reflection point on the road surface to the microphone position
    airAbsorptionCoefficients: np.ndarray
        1D array containing the air absorption coefficients computed at a set of frequency bands covering the 
        range from 0 to fs/2
    asphaltReflectionFilterTable: np.ndarray
        2D array containing the asphalt reflection filters computed at different reflection angles. The filters
        have a fixed number of taps ntaps = 10, and their impulse response at a set of angles from -89 to + 89 degrees
        is pre-computed and stored in this table in order to speed up the execution.
    simulation_params: dict
        Dictionary containing the keys:
        * `interp_method`: str, defines the interpolation method used in the delay line interpolated reads
        * `include_reflected_path`: bool, if `True` the received sample includes both the direct and the reflected sound,
        otherwise it includes just the direct sound
        * `include_air_abosrption`: bool, if `True` the air absorption is included in the simulation

    Methods
    ------
    initialize(src_pos: np.ndarray, mic_pos: np.ndarray):
        Sets the initial delays in the primary and secondary delay lines, according to the (initial) position of the
        source and the position of the microphone used for the current simulation
    update(src_pos: np.ndarray, mic_pos: np.ndarray, signal_sample: float):
        Updates delays on the primary and secondary delay lines given the new position of the source, and produces
        a new sample received by the microphone
    """

    def __init__(
            self,
            c: float,                       # speed of sound in air 
            fs: int,                        # sampling frequency
            Z0: int,                        # characteristic impedance of air
            road_material: Material,        # Absorption properties of road surface
            airAbsorptionCoefficients: np.ndarray,
            simulation_params: dict = {
                "interp_method": "Sinc",
                "include_reflected_path": True,
                "include_air_absorption": True,
            },
        ) -> None:
        """
        Creates a `SimulatorManager` object containing all the methods and attributes required to run the acoustic
        simulation. When the constructor is called, the environmental parameters previously defined in the object
        `Environment` are stored into the simulator manager (c, fs, Z0, road_material, airAbsorptionCoefficients). 
        In addition, the two `DelayLine` objects required for the simulation are instantiated as empty circular 
        arrays with 48000 entries. The `asphaltReflectionFilterTable` is pre-computed and stored as an instance 
        attribute containing the FIR filter coefficients that describe asphalt reflection at different angles of
        incidence of the sound waves. This will result in a reduction of the computational load,
        since the computation of a new filter depending on the instantaneous incidence angle is turned into a 
        table lookup. The set of considered incidence angles is stored in `_theta_vector` in degrees. In order to
        reduce the computational complexity, also the pseudo inverse matrix, needed by the least squares filter design
        method in the air absorption filters evaluation, is pre-computed. Finally, a set of auxiliary buffers are 
        instantiated and will be used for filtering purpose.

        Parameters
        ----------
        c : float
            Speed of sound in the air
        fs: int
            Sampling frequency used in the simulation
        Z0: float
            Characteristic impedance of the air at the chosen environmental conditions (i.e. T, p)
        road_material: Material
            Material object defining the absorption and reflection properties of the road surface
        airAbsorptionCoefficients: np.ndarray
            1D Array containing the air absorption coefficients computed in the defined frequency bands at the chosen 
            environmental conditions (i.e. T, p, rel. humidity)
        
        Raises
        ------
        KeyError:
            If simulation_params does not contain keys `interp_method`, `include_reflected_path` and 
            `include_air_absorption`.
        """
        self.c = c
        self.fs = fs
        self.Z0 = Z0
        self.road_material = road_material
        self.airAbsorptionCoefficients = airAbsorptionCoefficients

        # Define simulation parameters
        if ("interp_method" not in simulation_params or "include_reflected_path" 
            not in simulation_params or "include_air_absorption" not in simulation_params):
            raise KeyError("simulation_params must be a dict with keys 'interp_method' (string),"
              "'include_reflection_path' (bool) and 'include_air_absorption (bool'")
        self.simulation_params = simulation_params

        # Instantiation of two delay lines used for the simulations
        self.primaryDelLine = DelayLine(N = 48000, num_read_ptrs = 2, 
            interpolation = self.simulation_params["interp_method"]) 
        self.secondaryDelLine = DelayLine(N = 48000, num_read_ptrs = 1, 
            interpolation = self.simulation_params["interp_method"])

        # Array containing all possible incidence angles for pre-computed reflection filter table
        self._theta_vector = np.arange(-89,90,1)

        # Instantiation of table containing asphalt reflection filters
        self.asphaltReflectionFilterTable = self._compute_angle_reflection_table(ntaps = 11)
       
        # Buffers to store previous data read from delay lines for filtering purpose
        self._read1Buf = deque(np.zeros(11), maxlen=11)
        self._read2Buf = deque(np.zeros(11), maxlen=11)
        self._read3Buf = deque(np.zeros(11), maxlen=11)
        self._read4Buf = deque(np.zeros(11), maxlen=11)

        # Compute air absorption filter 
        freqs = np.linspace(0, self.fs / 2, len(airAbsorptionCoefficients))
        w = 2 * np.pi * freqs / (fs / 2)
        
        # Filter order
        L = 10

        # Matrix A
        A = np.zeros((len(w),int(L/2 + 1)))
        A[:,0] = 1
        for i in range(len(w)):
            for j in range(1,int(L/2 + 1)):
                A[i,j] = 2 * math.cos(w[i] * j)
        
        self._pseud_A = np.linalg.inv(A.T.dot(A)).dot(A.T)
        

    def initialize(self, src_pos: np.ndarray, mic_pos: np.ndarray) -> None:
        """
        Computes the propagation delays along the paths:
        * Direct path from source initial position to microphone
        * Path from image source position to incidence point on asphalt surface, along reflected path
        * Path from incidence point on asphalt surface to microphone, along reflected path
        Sets the corresponding initial delays on the two delay lines.

        This method must be called at the start of every simulation to set the initial conditions that
        ensure no discontinuity is generated at the first simulation instant.

        Parameters
        ----------
        src_pos : np.ndarray
            1D array containing the cartesian coordinates [x,y,z] denoting the initial position of the sound source
        mic_pos : np.ndarray
            1D array containing the cartesian coordinates [x,y,z] denoting the position of the microphone
        """

        # Compute direct distance and delay
        _, tau = self._compute_delay(src_pos, mic_pos)
        
        # Compute incidence angle
        theta = self._compute_angle(src_pos, mic_pos)

        # Compute distance and delay between src and reflection point
        a = src_pos[2] / math.sin(theta)
        tau_1 = a / self.c
        
        # Compute distance and delay between reflection point and microphone
        b = src_pos[2] / math.sin(theta)
        tau_2 = b / self.c

        # Set initial delays
        self.primaryDelLine.set_delays(np.array([tau * self.fs, tau_1 * self.fs]))
        self.secondaryDelLine.set_delays(np.array([tau_2 * self.fs]))
        
    def update(self, src_pos: np.ndarray, mic_pos: np.ndarray, signal_sample: float) -> float:
        """
        This method is used to compute the output of one instant of the simulation. It takes as inputs the instantaneous
        positions of the source and of the microphone used in the current simulation, and the sample of the source
        signal emitted at the considered instant. It then performs the update of the `DelayLine` objects, writing the new 
        sample `signal_sample` on the primary line, and computes the relative distances and delays between the source,
        the road surface at the incidence point, and the microphone, together with the incidence angle. According to
        these delays, the `DelayLine` read pointers positions are updated. Then, the output sample received at the
        microphone position is computed as a sum of two components:
        * the direct component of the sound field, filtered with by air absorption filter ,depending on the distance
        between the source and the microphone, and attenuated depending on the travelled distance;
        * the reflected component: the reflected sound is analyzed using a cascade of 3 filters and two multiplications.
        The reflected sound wave travels from the source to the road surface, along a trajectory computed 
        using the principles of geometrical acoustics. This path is represented by a propagation delay, a filter that
        models the air absorption depending on the distance between the source and the incidence point, and an
        attenuation factor depending on the same travelled distance. At the incidence point, the sound wave is 
        filtered with a second FIR filter representing the asphalt reflection, depending on the road surface material 
        and on the incidence angle. After the reflection, the sound wave travels from the reflection point to the 
        microphone, along a path modeled by a second delay element, a second air absorption filter depending on the
        distance between reflection point and microphone, and a second attenuation factor depending on the same
        distance. 

        The two components are summed to produce the output sample, that is returned by the update function. All the 
        filtering operations are performed in the time domain, as the convolution between the considered signals 
        and the FIR filters.

        Parameters
        ----------
        src_pos : np.ndarray
            1D array containing the cartesian coordinates [x,y,z] denoting the position of the sound source
        mic_pos : np.ndarray
            1D array containing the cartesian coordinates [x,y,z] denoting the position of the microphone
        signal_sample : float
            Sample produced by the source at the considered instant

        Returns
        -------
        float
            Sample of the signal recorded by the cosidered microphone at the given time instant
        """

        # Compute direct distance and delay
        d, tau = self._compute_delay(src_pos, mic_pos)

        # Compute incidence angle
        theta = self._compute_angle(src_pos, mic_pos)

        # Compute distance and delay between src and reflection point
        a = src_pos[2] / math.sin(theta)
        tau_1 = a / self.c
        
        # Compute distance between reflection point and microphone
        b = mic_pos[2] / math.sin(theta)
        tau_2 = b / self.c

        # Update delays and get new sample reads
        y_primary = self.primaryDelLine.update_delay_line(signal_sample, np.array([tau * self.fs, 
            tau_1 * self.fs]))

        # Store the read samples in a circular array for filtering with air absorption and asphalt reflection
        self._read1Buf.appendleft(y_primary[0])
        self._read2Buf.appendleft(y_primary[1])
        
        y_received = 0
        ### DIRECT PATH ###

        if self.simulation_params["include_air_absorption"]:
            # Attenuation due to air absorption
            filt_coeffs = self._compute_air_absorption_filter(d, numtaps = 11)
            sample_eval = filt_coeffs.dot(list(self._read1Buf))
        else:
            sample_eval = self._read1Buf[0]
        
        # Attenuation due to distance
        att = self._compute_sound_attenuation(d)

        # Direct Path Output Sample
        y_dir =  att * sample_eval

        ### REFLECTED PATH ###
        if self.simulation_params["include_reflected_path"]:
        # 1. From Source to Road Surface

            if self.simulation_params["include_air_absorption"]:

                # Attenuation due to air absorption
                filt_coeffs = self._compute_air_absorption_filter(a, numtaps = 11)
                sample_eval = filt_coeffs.dot(list(self._read2Buf))
            else:
                sample_eval = self._read2Buf[0]
            
            # Attenuation due to distance
            att = self._compute_sound_attenuation(a)
            sample_eval = att * sample_eval
            
            # Write output of first filter in buffer, to be able to apply second filter in cascade
            self._read3Buf.appendleft(sample_eval)

            # 2. Asphalt Absorption
            asphalt_filter_coeffs = self._get_asphalt_reflection_filter(90 - math.degrees(theta))
            
            sample_eval = asphalt_filter_coeffs.dot(list(self._read3Buf))
            
            
            # 3. Second path in air --> Secondary Delay Line
            y_secondary = self.secondaryDelLine.update_delay_line(sample_eval, np.array([tau_2 * self.fs]))

            # Store output of secondary delay line in buffer for cascade air abs filter
            self._read4Buf.appendleft(y_secondary[0])

            # 4. From Road Surface to Receiver
            if self.simulation_params["include_air_absorption"]:
                # Attenuation due to air absorption
                filt_coeffs = self._compute_air_absorption_filter(b, numtaps = 11)
                sample_eval = filt_coeffs.dot(list(self._read4Buf))
            else:
                sample_eval = self._read4Buf[0]
            
            # Attenuation due to distance
            att = self._compute_sound_attenuation(b)

            y_refl = att * sample_eval

            # Compute received sample as sum of direct and reflected path outputs
            y_received = y_refl
        
        y_received = y_received + y_dir

        return y_received
    
    def _compute_air_absorption_filter(self, distance: float, numtaps: int) -> np.ndarray:
        """
        Computes air absorption filter as a FIR filter with `numtaps` coefficients. The filter depends
        on the distance between the source and the receiver.

        Parameters
        ----------
        distance : float
            Distance travelled by the sound wave, in meters
        numtaps : int
            Number of coefficients of the FIR filter

        Returns
        -------
        np.ndarray
            1D array containing `numtaps` FIR filter coefficients modelling air absorption
        """
        filt_coeffs = np.empty(numtaps, np.float64)
        alpha = 10 ** (-self.airAbsorptionCoefficients * distance / 20)     # Convert coeffs in dB to linear scale
        filt_coeffs[int((numtaps+1)/2)-1:] = self._pseud_A.dot(alpha)
        filt_coeffs[0:int((numtaps+1)/2)-1] = np.flip(filt_coeffs[int((numtaps+1)/2):])
        
        return filt_coeffs

    def _compute_angle_reflection_table(self, ntaps: int = 11) -> np.ndarray:
        """
        Computes the coefficients of the FIR filters modelling asphalt reflection, at a set of different
        incidence angles, and stores them in a table. Each row of the table contains the `ntaps` FIR filter 
        corresponding to a different incidence angle. The set of angles is defined in `self._theta_vector`.

        The filter coefficients are computed using the `scipy.signal.firwin2` method, given the reflection coefficients
        computed in the frequency bands `center_freqs`, defined in the instance attribute `road_material`. 
        The reflection coefficients are derived from the  absorption coefficients of the road surface, defined
        in the instance attribute `road_material`. The road surface is assumed to be  locally reactive, meaning 
        that its impedance Z does not depend on the incidence angle.

        Parameters
        ----------
        ntaps : int
            Number of FIR filter coefficients, must be odd

        Returns
        -------
        np.ndarray
            2D Array with len(self._theta_vector) rows, each containing `ntaps` FIR filter coefficients corresponding
            to each angle defined in `self._theta_vector`
        """
        
        Z0 = self.Z0
        refl = np.sqrt(1 - np.array(self.road_material.absorption["coeffs"]))
        
        # The impedance does not depend on the angle of incidence -> locally reactive surface
        Z = - Z0 * (refl + 1) / (refl-1)
        
        b_fir = np.zeros((len(self._theta_vector), ntaps))
        
        # Compute filters coefficients for all thetas
        for idx, theta in enumerate(self._theta_vector):
            R = np.abs((Z * np.cos(math.radians(theta)) - Z0) / (Z * np.cos(math.radians(theta)) + Z0))
            b_fir[idx] = scipy.signal.firwin2(ntaps, np.array(self.road_material.absorption["center_freqs"]) 
                / (self.fs/2), R)
        
        return b_fir
    
    def _get_asphalt_reflection_filter(self, theta: float) -> np.ndarray:
        """
        Select the asphalt reflection filter from the pre-computed table given the incidence angle of the 
        sound wave hitting the road surface, in degrees. For the table lookup, the incidence angle is 
        rounded to the closest integer in the range defined in `self._theta_vector` attribute.

        Parameters
        ----------
        theta : float
            Incidence angle of the sound wave hitting the road surface, in degrees

        Returns
        -------
        np.ndarray
            1D Array containing asphalt reflection filter coefficients computed at angle theta
        """

        idx = np.where(self._theta_vector == round(theta))
        idx = idx[0][0]
        return self.asphaltReflectionFilterTable[idx]

    def _compute_sound_attenuation(self, distance: float) -> float:
        """
        Computes the attenuation of sound due to its propagation in the empty space. This attenuation depends
        only on the distance between the source and the receiver and is computed assuming a spherical propagation
        model.

        Parameters
        ----------
        distance : float
            Relative distance between the sound source and the receiver

        Returns
        -------
        float
            Attenuation factor
        """
        return 1 / (distance)
    
    def _compute_delay(self, src_pos: np.ndarray, mic_pos: np.ndarray) -> tuple[float, float]:
        """
        Computes the distance between the source and the microphone, and the time needed by the sound to travel 
        from source to microphone, given their distance and the speed of sound in air.

        Parameters
        ----------
        src_pos : np.ndarray
            1D array containing the cartesian coordinates [x,y,z] denoting the position of the source
        mic_pos : np.ndarray
            1D array containing the cartesian coordinates [x,y,z] denoting the position of the microphone

        Returns
        -------
        Tuple[float, float]
            Tuple containing the distance between source and microphone (in meters) and the time delay (in seconds)
        """

        d = math.sqrt(np.sum((src_pos - mic_pos) ** 2))
        tau = d / self.c
        return d, tau
    
    def _compute_angle(self, src_pos: np.ndarray, mic_pos: np.ndarray) -> float:
        """
        Computes the incidence angle of the sound wave travelling from the source to the road surface, along the
        reflected path connecting the source to the microphone with a single reflection generated by the road.

        Parameters
        ----------
        src_pos : np.ndarray
            1D array containing the cartesian coordinates [x,y,z] denoting the position of the source
        mic_pos : np.ndarray
            1D array containing the cartesian coordinates [x,y,z] denoting the position of the microphone

        Returns
        -------
        float
            Incidence angle of sound wave travelling from source to road surface along reflected path, in radians
        """

        # Distance between image and microphone
        dist = math.sqrt(np.sum(((src_pos - np.array([0, 0 , 2*src_pos[2]])) - mic_pos) ** 2))
        # Incidence angle
        theta = math.asin((src_pos[2] + mic_pos[2]) / dist)
        return theta


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyroadacoustics",
    version="1.0.2",
    author="Stefano Damiano",
    author_email="stefano.damiano@esat.kuleuven.be",
    description="A package for simulating the sound propagation in a road scenario",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/steDamiano/pyroadacoustics",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Information Technology",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Multimedia :: Sound/Audio :: Analysis",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    packages = ["pyroadacoustics"],
    python_requires=">=3.7",
    package_data= {
        "" : ['*.json']
    },
    install_requires=[
        "numpy",
        "scipy>=0.18.0",
    ],
    keywords="outdoor aocustics automotive signal processing traffic sound simulation",
)